# coding=utf-8

from __future__ import absolute_import, print_function

from unittest import TestCase, skip

from abilian.sbe.apps.documents.models import Document, Folder, icon_for


def check_editable(object):
    if hasattr(object, '__editable__'):
        for k in object.__editable__:
            assert hasattr(object, k)


def test_title_prevails():
    f = Folder(name=u'name', title=u'title')
    assert f.title == u'title'
    assert f.name == u'title'

    f = Folder(name=u'name', title=None)
    assert f.title == u'name'
    assert f.name == u'name'

    f = Folder(name=u'name')
    assert f.title == u'name'
    assert f.name == u'name'


def test_folder_editables():
    root = Folder(title="/")
    check_editable(root)


def test_folder_can_create_documents():
    root = Folder(title="/")

    document = root.create_document("doc")

    assert len(root.children) == 1
    assert document.title == "doc"
    assert document.path == "/doc"
    assert document in root.documents
    assert document in root.children
    assert document.parent == root


def test_folder_can_create_subfolders():
    root = Folder(title="/")

    subfolder = root.create_subfolder("folder")
    assert len(root.children) == 1
    assert subfolder.title == "folder"
    assert subfolder.path == "/folder"
    assert subfolder in root.subfolders
    assert subfolder in root.children
    assert subfolder.parent == root


def test_nested_subobjects():
    root = Folder(title="/")
    subfolder = root.create_subfolder("folder1")
    subsubfolder = subfolder.create_subfolder("folder2")
    document = subfolder.create_document("doc")

    assert len(root.children) == 1
    assert len(subfolder.children) == 2
    assert subsubfolder.title == "folder2"
    assert subsubfolder.path == "/folder1/folder2"
    assert document.title == "doc"
    assert document.path == "/folder1/doc"

    assert root.get_object_by_path("/folder1") == subfolder
    assert root.get_object_by_path("/folder1/folder2") == subsubfolder
    assert root.get_object_by_path("/folder1/doc") == document


def test_folder_is_clonable():
    root = Folder(title="/")
    clone = root.clone()

    assert clone.title == root.title
    assert clone.path == root.path


# FIXME: skipped for now as these tests break other tests down the line (in wiki).
@skip
class TestDocument(TestCase):

    def test_document_editables(self):
        doc = Document()
        check_editable(doc)

    def test_content_length(self):
        doc = Document()
        doc.set_content("tototiti", "application/binary")
        assert doc.content_length == len("tototiti")

    def test_document_is_clonable(self):
        root = Folder(title="/")
        doc = root.create_document(title="toto")
        doc.content = "tototiti"

        clone = doc.clone()
        assert clone.title == doc.title

    def test_document_has_an_icon(self):
        root = Folder(title="/")
        doc = root.create_document(title="toto")
        doc.content_type = "image/jpeg"
        filename = doc.icon.split("/")[-1]
        assert filename in ("jpg.png", "jpeg.png"), doc.icon

    def test_icon_from_mime_type(self):
        icon = icon_for("text/html")
        filename = icon.split("/")[-1]
        assert filename in ("html.png", "htm.png"), icon



# coding=utf-8
"""
Celery tasks related to document transformation and preview.
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging
from contextlib import contextmanager

from celery import shared_task

from abilian.core.extensions import db
from abilian.services import converter, get_service
from abilian.services.conversion import ConversionError, HandlerNotFound

logger = logging.getLogger(__package__)


@contextmanager
def get_document(document_id, session=None):
    """Context manager that yields (session, document).
    """
    from .models import Document

    doc_session = session
    if session is None:
        doc_session = db.create_scoped_session()

    with doc_session.begin_nested():
        query = doc_session.query(Document)
        document = query.get(document_id)
        yield (doc_session, document)

    # cleanup
    if session is None:
        doc_session.commit()
        doc_session.close()


@shared_task
def process_document(document_id):
    """Run document processing chain.
    """
    with get_document(document_id) as (session, document):
        if document is None:
            return

        # True = Ok, None means no check performed (no antivirus present)
        is_clean = _run_antivirus(document)
        if is_clean is False:
            return

    preview_document.delay(document_id)
    convert_document_content.delay(document_id)


def _run_antivirus(document):
    antivirus = get_service('antivirus')
    if antivirus and antivirus.running:
        is_clean = antivirus.scan(document.content_blob)
        if 'antivirus_task' in document.content_blob.meta:
            del document.content_blob.meta['antivirus_task']
        return is_clean
    return None


@shared_task
def antivirus_scan(document_id):
    """Return antivirus.scan() result
    """
    with get_document(document_id) as (session, document):
        if document is None:
            return
        return _run_antivirus(document)


@shared_task
def preview_document(document_id):
    """Compute the document preview images with its default preview size.
    """
    with get_document(document_id) as (session, document):
        if document is None:
            # deleted after task queued, but before task run
            return

        try:
            converter.to_image(document.content_digest, document.content,
                               document.content_type, 0, document.preview_size)
        except ConversionError as e:
            logger.info(
                'Preview failed: %s',
                str(e),
                exc_info=True,
                extra={'stack': True})


@shared_task
def convert_document_content(document_id):
    """Convert document content.
    """
    with get_document(document_id) as (session, doc):
        if doc is None:
            # deleted after task queued, but before task run
            return

        error_kwargs = dict(exc_info=True, extra={'stack': True})

        conversion_args = (doc.content_digest, doc.content, doc.content_type)

        if doc.content_type == "application/pdf":
            doc.pdf = doc.content
        else:
            try:
                doc.pdf = converter.to_pdf(*conversion_args)
            except HandlerNotFound as e:
                doc.pdf = b""
            except ConversionError as e:
                doc.pdf = b""
                logger.info("Conversion to PDF failed for document %s: %s",
                            doc.name, e, **error_kwargs)

        try:
            doc.text = converter.to_text(doc.content_digest, doc.content,
                                         doc.content_type)
        except ConversionError as e:
            doc.text = ""
            logger.info("Conversion to text failed for document %s: %s",
                        doc.name, e, **error_kwargs)

        doc.extra_metadata = {}
        try:
            doc.extra_metadata = converter.get_metadata(*conversion_args)
        except ConversionError as e:
            logger.warning("Metadata extraction failed on document %s: %s",
                           doc.name, e, **error_kwargs)
        except UnicodeDecodeError as e:
            logger.error("Unicode issue on document %s: %s", doc.name, e,
                         **error_kwargs)
        except Exception as e:
            logger.error("Other issue on document %s: %s", doc.name, e,
                         **error_kwargs)

        if doc.text:
            import langid
            doc.language = langid.classify(doc.text)[0]

        doc.page_num = doc.extra_metadata.get("PDF:Pages", 1)



from __future__ import absolute_import, print_function


class Env(dict):
    """
    A dictionnary whose objects can be accessed by '.' notation, saving a few
    keystrokes.
    """

    def __init__(self, label=None, **kw):
        # self.bread_crumbs = self.breadcrumbs = make_bread_crumbs(label=label)
        for key, value in kw.items():
            self[key] = value

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, key):
        return self[key]



# Note: this test suite is using pytest instead of the unittest-based scaffolding
# provided by SBE. Hopefully one day all of SBE will follow.

from __future__ import absolute_import, print_function

from datetime import datetime, timedelta

import pytest

import abilian.i18n
from abilian.core.signals import activity
from abilian.sbe.app import create_app
from abilian.sbe.apps.communities.common import activity_time_format


@pytest.yield_fixture
def app():
    app = create_app()

    # We need some incantations here to make babel work in the test
    babel = abilian.i18n.babel
    babel.locale_selector_func = None

    yield app

    # Signals are globals and apparently need to be cleaned up
    # At this point, only the "activity" signal seems to have a side effect.
    activity._clear_state()


def test_activity_time_format(app):

    # We need the app context because of Babel.
    with app.app_context():
        then = datetime(2017, 1, 1, 12, 0, 0)

        now = then + timedelta(0, 5)
        assert activity_time_format(then, now) == '5s'

        now = then + timedelta(0, 5 * 60)
        assert activity_time_format(then, now) == '5m'

        now = then + timedelta(0, 5 * 60 * 60)
        assert activity_time_format(then, now) == '5h'

        now = then + timedelta(1, 5)
        assert activity_time_format(then, now) == '1d'

        now = then + timedelta(60, 5)
        assert activity_time_format(then, now) == 'Jan 1'

        now = then + timedelta(365 + 60, 5)
        assert activity_time_format(then, now) == 'Jan 2017'



# coding=utf-8
"""
Notifications
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe

# Constants
TOKEN_SERIALIZER_NAME = "unsubscribe_sbe"


def register_plugin(app):
    cfg = app.config.setdefault('ABILIAN_SBE', {})
    cfg.setdefault('DAILY_SOCIAL_DIGEST_SUBJECT',
                   u'Des nouvelles de vos communautés')
    sbe.init_app(app)

    # TODO: Slightly confusing. Reorg?
    from .views import notifications, social  # noqa
    from .tasks.social import DIGEST_TASK_NAME, DEFAULT_DIGEST_SCHEDULE

    CELERYBEAT_SCHEDULE = app.config.setdefault('CELERYBEAT_SCHEDULE', {})

    if DIGEST_TASK_NAME not in CELERYBEAT_SCHEDULE:
        CELERYBEAT_SCHEDULE[DIGEST_TASK_NAME] = DEFAULT_DIGEST_SCHEDULE

    app.register_blueprint(notifications)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

import bleach
from wtforms import StringField, TextAreaField
from wtforms.fields.html5 import URLField

from abilian.i18n import _l
from abilian.web.forms import Form, ValidationError
from abilian.web.forms.fields import DateTimeField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required
from abilian.web.forms.widgets import RichTextWidget

ALLOWED_TAGS = [
    'a',
    'abbr',
    'acronym',
    'b',
    'blockquote',
    'br',
    'code',
    'em',
    'i',
    'li',
    'ol',
    'strong',
    'ul',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'p',
    'u',
    'img',
]

ALLOWED_ATTRIBUTES = {
    '*': ['title'],
    'p': ['style'],
    'a': ['href', 'title'],
    'abbr': ['title'],
    'acronym': ['title'],
    'img': ['src', 'alt', 'title'],
}

ALLOWED_STYLES = ['text-align',]

WIDGET_ALLOWED = {}
for attr in ALLOWED_TAGS:
    allowed = ALLOWED_ATTRIBUTES.get(attr, True)
    if not isinstance(allowed, bool):
        allowed = {tag: True for tag in allowed}
    WIDGET_ALLOWED[attr] = allowed


class EventForm(Form):
    title = StringField(
        label=_l("Title"), filters=(strip,), validators=[required()])

    start = DateTimeField(_l("Start"), validators=[required()])
    end = DateTimeField(_l("End"), validators=[required()])

    location = TextAreaField(label=_l("Location"), filters=(strip,))

    url = URLField(label=_l("URL"), filters=(strip,))

    description = TextAreaField(
        label=_l("Description"),
        widget=RichTextWidget(allowed_tags=WIDGET_ALLOWED),
        filters=(strip,),
        validators=[required()])

    def validate_description(self, field):
        field.data = bleach.clean(
            field.data,
            tags=ALLOWED_TAGS,
            attributes=ALLOWED_ATTRIBUTES,
            styles=ALLOWED_STYLES,
            strip=True)

    def validate_end(self, field):
        if self.start.data > self.end.data:
            raise ValidationError(_l("End date/time must be after start"))


EventForm.start.kwargs['raw_data'] = [" | 09:00"]
EventForm.end.kwargs['raw_data'] = [" | 18:00"]



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function

import logging
import uuid
from pathlib import Path

import pkg_resources

from abilian.core.util import fqcn

logger = logging.getLogger(__name__)

STATIC_DIR = pkg_resources.resource_filename(__name__, 'static')
LESSCSS_FILE = str(Path(STATIC_DIR, 'less', 'abilian-sbe.less'))
JS = (
    'js/sbe-datatable.js',
    'js/jquery.fileapi.js',
    'js/folder.js',
    'js/folder_edit.js',
    'js/folder_upload.js',
    'js/document_viewer.js',
    'js/bootstrap-tagsinput.js',)


class AbilianSBE(object):
    """
    Base extension required by abilian.sbe.apps
    """

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        # False: it's ok if antivirus task was run but service couldn't get a result
        app.config.setdefault('ANTIVIRUS_CHECK_REQUIRED', False)
        app.config.setdefault('SBE_FORUM_REPLY_BY_MAIL', False)

        if FQCN in app.extensions:
            return

        app.extensions[FQCN] = self

        # register i18n
        app.extensions['babel'].add_translations('abilian.sbe')

        # sbe static assets
        app.add_static_url(
            'abilian/sbe', STATIC_DIR, endpoint='abilian_sbe_static')
        app.extensions['webassets'].append_path(
            STATIC_DIR, app.static_url_path + '/abilian/sbe')

        app.register_asset('js', *JS)
        app.register_asset('css', LESSCSS_FILE)

        # Jinja
        logger.info('Register jinja context processors')
        app.context_processor(inject_template_utils)


FQCN = fqcn(AbilianSBE)
sbe = AbilianSBE()


def inject_template_utils():
    return dict(uuid=uuid.uuid1)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime, timedelta
from email.parser import FeedParser
from pathlib import Path
from unittest import TestCase

import pytest
from flask import url_for
from mock import Mock, patch
from six import text_type

from abilian.core.models.subjects import User
from abilian.sbe.apps.communities.models import MANAGER, MEMBER
from abilian.sbe.apps.communities.tests.base import CommunityBaseTestCase, \
    CommunityIndexingTestCase
from abilian.sbe.testing import BaseTestCase

from ..commands import inject_email
from ..models import Post, Thread, ThreadClosedError
from ..tasks import build_reply_email_address, extract_email_destination, \
    process


class Test(TestCase):

    def test_create_post(self):
        thread = Thread(title="Test thread")
        post = thread.create_post()
        assert post in thread.posts
        assert post.name == 'Test thread'

        thread.title = 'new title'
        assert thread.name == 'new title'
        assert post.name == 'new title'

    def test_closed_property(self):
        thread = Thread(title='Test Thread')
        assert thread.closed is False
        thread.closed = True
        assert thread.closed is True
        thread.closed = 0
        assert thread.closed is False
        thread.closed = 1
        assert thread.closed is True
        assert thread.meta['abilian.sbe.forum']['closed'] is True

    def test_thread_closed_guard(self):
        thread = Thread(title='Test Thread')
        thread.create_post()
        thread.closed = True

        with pytest.raises(ThreadClosedError):
            thread.create_post()

        p = Post(body_html='ok')

        with pytest.raises(ThreadClosedError):
            p.thread = thread

        thread.closed = False
        p.thread = thread
        assert len(thread.posts) == 2

        thread.closed = True
        with pytest.raises(ThreadClosedError):
            del thread.posts[0]

        assert len(thread.posts) == 2

        with pytest.raises(ThreadClosedError):
            # actually thread.posts will be replaced by `[]` and we can't prevent
            # this, but exception has been raised
            thread.posts = []

        thread.closed = False
        thread.posts = [p]
        thread.closed = True
        assert thread.posts == [p]
        assert p.thread is thread
        with pytest.raises(ThreadClosedError):
            p.thread = None

    def test_change_thread_copy_name(self):
        thread = Thread(title='thread 1')
        thread2 = Thread(title='thread 2')
        post = Post(thread=thread, body_html='post content')
        assert post.name == thread.name

        post.thread = thread2
        assert post.name == thread2.name

    def test_task_process_email(self):
        """Test the process_email function."""

        marker = '_____Write above this line to post_____'

        message = get_email_message_from_file('reply.email')
        newpost = process(message, marker)[0]
        assert newpost

        message = get_email_message_from_file('reply_nocharset_specified.email')
        newpost = process(message, marker)[0]
        assert newpost

        message = get_email_message_from_file('reply_no_marker.email')
        self.assertRaises(LookupError, process, message, marker)

        # dubious check
        message = get_email_message_from_file('reply_no_textpart.email')
        self.assertRaises(LookupError, process, message, marker)


class IndexingTestCase(CommunityIndexingTestCase):

    def test_thread_indexed(self):
        thread = Thread(title='Community 1', community=self.community)
        self.session.add(thread)
        thread_other = Thread(title='Community 2: other', community=self.c2)
        self.session.add(thread_other)
        self.session.commit()

        svc = self.svc
        obj_types = (Thread.entity_type,)
        with self.login(self.user_no_community):
            res = svc.search('community', object_types=obj_types)
            assert len(res) == 0

        with self.login(self.user):
            res = svc.search('community', object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit['object_key'] == thread.object_key

        with self.login(self.user_c2):
            res = svc.search('community', object_types=obj_types)
            assert len(res) == 1
            hit = res[0]
            assert hit['object_key'] == thread_other.object_key


class NoLoginViewTest(CommunityBaseTestCase):
    """
    Test correct url response, without login or security involved.
    """

    def test(self):
        response = self.client.get(
            url_for("forum.index", community_id=self.community.slug))
        self.assert200(response)


class ViewTestCase(CommunityBaseTestCase):
    no_login = False
    SERVICES = ('security',)

    def test_posts_ordering(self):
        thread = Thread(community=self.community, title='test ordering')
        self.session.add(thread)
        t1 = datetime(2014, 6, 20, 15, 0, 0)
        p1 = Post(thread=thread, body_html='post 1', created_at=t1)
        t2 = datetime(2014, 6, 20, 15, 1, 0)
        p2 = Post(thread=thread, body_html='post 2', created_at=t2)
        self.session.flush()
        p1_id, p2_id = p1.id, p2.id
        assert [p.id for p in thread.posts] == [p1_id, p2_id]

        # set post1 created after post2
        t1 = t1 + timedelta(minutes=2)
        p1.created_at = t1
        self.session.flush()
        self.session.expire(thread)  # force thread.posts refreshed from DB
        assert [p.id for p in thread.posts] == [p2_id, p1_id]

    def test_create_thread_and_post(self):
        # activate email reply
        self.app.config['SBE_FORUM_REPLY_BY_MAIL'] = True
        self.app.config['MAIL_ADDRESS_TAG_CHAR'] = '+'

        # create a new user, add him/her to the current community as a MANAGER
        self.user = User(
            email='user_1@example.com', password='azerty', can_login=True)
        self.session.add(self.user)
        self.community.set_membership(self.user, MANAGER)
        self.session.commit()
        self.client_login(self.user.email, self.user.password)

        mail = self.app.extensions['mail']
        with mail.record_messages() as outbox:
            title = "Brand new thread"
            content = "shiny thread message"
            url = url_for("forum.new_thread", community_id=self.community.slug)
            data = dict(title=title, message=content)
            data['__action'] = "create"
            data['send_by_email'] = "y"
            response = self.client.post(url, data=data)
            self.assertStatus(response, 302)

            # extract the thread_id from the redirection url in response
            threadid = response.location.rsplit('/', 2)[1]

            # retrieve the new thread, make sur it has the message
            url = url_for(
                "forum.thread",
                thread_id=threadid,
                community_id=self.community.slug,
                title=title)
            response = self.client.get(url)
            self.assertStatus(response, 200)
            self.assertIn(content, response.data.decode("utf-8"))

            # check the email was sent with the new thread
            assert len(outbox) == 1
            assert outbox[0].subject == '[My Community] Brand new thread'

        # reset the outbox for checking threadpost email
        with mail.record_messages() as outbox:
            content = data['message'] = "my cherished post"
            del data['title']
            response = self.client.post(url, data=data)
            self.assertStatus(response, 302)

            # retrieve the new thread, make sur it has the message
            url = url_for(
                "forum.thread",
                thread_id=threadid,
                community_id=self.community.slug,
                title=title)
            response = self.client.get(url)
            self.assertStatus(response, 200)
            self.assertIn(content, response.data.decode("utf-8"))

            # check the email was sent with the new threadpost
            assert len(outbox) == 1
            expected = '[My Community] Brand new thread'
            assert text_type(outbox[0].subject) == expected

    def test_create_thread_informative(self):
        """
        Test with 'informative' community. No mail sent, unless user is MANAGER
        """
        assert self.community.type == 'informative'
        # create a new user, add him/her to the current community
        self.user = User(
            email='user_1@example.com', password='azerty', can_login=True)
        self.session.add(self.user)
        self.community.set_membership(self.user, MEMBER)
        self.session.commit()

        title = "Brand new thread"
        content = "shiny thread message"
        url = url_for("forum.new_thread", community_id=self.community.slug)
        data = dict(title=title, message=content)
        data['__action'] = "create"

        mail = self.app.extensions['mail']
        self.client_login(self.user.email, self.user.password)

        with mail.record_messages() as outbox:
            data['send_by_email'] = "y"  # actually should not be in html form
            response = self.client.post(url, data=data)
            self.assertStatus(response, 302)
            assert len(outbox) == 0

        self.community.set_membership(self.user, MANAGER)
        self.session.commit()

        with mail.record_messages() as outbox:
            data['send_by_email'] = "y"  # should be in html form
            response = self.client.post(url, data=data)
            self.assertStatus(response, 302)
            assert len(outbox) == 1

        with mail.record_messages() as outbox:
            del data['send_by_email']
            response = self.client.post(url, data=data)
            self.assertStatus(response, 302)
            assert len(outbox) == 0

        self.client_logout()


def get_string_from_file(filename='notification.email'):
    """Load a test email, return as string."""
    filepath = Path(__file__).parent / 'data' / filename
    with filepath.open('rt', encoding='utf-8') as email_file:
        email_string = email_file.read()
    return email_string


def get_email_message_from_file(filename='notification.email'):
    """Load a mail and parse it into a email.message."""
    email_string = get_string_from_file(filename)
    parser = FeedParser()
    parser.feed(email_string)
    email_message = parser.close()
    return email_message


class CommandsTest(TestCase):

    @patch('fileinput.input')
    @patch('abilian.sbe.apps.forum.commands.process_email')
    def test_parse_forum_email(self, mock_process_email, mock_email):
        """
        No processing is tested only parsing into a email.message
        and verifying inject_email() logic
        """
        # first load a test email returned by the mock_email
        mock_email.return_value = get_string_from_file('notification.email')

        # test the parsing function
        inject_email()

        # assert the email is read
        assert mock_email.called
        # assert a call on the celery task was made implying a message creation
        assert mock_process_email.delay.called

        ##
        mock_email.reset_mock()
        mock_process_email.delay.reset_mock()
        assert not mock_email.called
        assert not mock_process_email.delay.called

        mock_email.return_value = get_string_from_file('defects.email')
        inject_email()
        assert mock_email.called
        assert not mock_process_email.delay.called


class TasksTest(BaseTestCase):

    def test_build_reply_email_address(self):
        expected_reply_address = 'test+P-fr-3-4-5a6f41c0e7d916b6f15992d7207e47b2@testcase.app.tld'
        self.app.config['MAIL_ADDRESS_TAG_CHAR'] = '+'
        post = Mock()
        post.id = 2
        post.thread_id = 3
        member = Mock()
        member.id = 4
        with self.app.test_request_context(
                '/build_reply_email_address',
                headers=[('Accept-Language', 'fr')]):
            replyto = build_reply_email_address('test', post, member,
                                                'testcase.app.tld')
        self.assertIn(expected_reply_address, replyto)

    def test_extract_mail_destination(self):
        self.app.config['MAIL_ADDRESS_TAG_CHAR'] = '+'
        self.app.config['MAIL_SENDER'] = 'test@testcase.app.tld'
        test_address = 'test+test+P-fr-3-4-5a6f41c0e7d916b6f15992d7207e47b2@testcase.app.tld'
        infos = extract_email_destination(test_address)
        self.assertIn('fr', infos[0])  # locale
        self.assertIn('3', infos[1])  # thread_id
        self.assertIn('4', infos[2])  # post.id



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function

import hashlib
import logging
from collections import Counter
from datetime import datetime
from functools import wraps
from io import BytesIO
from operator import attrgetter
from pathlib import Path
from time import gmtime, strftime

import openpyxl
import pytz
import sqlalchemy as sa
from flask import current_app, flash, g, jsonify, redirect, render_template, \
    request, session, url_for
from flask_login import current_user, login_required
from openpyxl.writer.write_only import WriteOnlyCell
from six import text_type
from werkzeug.exceptions import BadRequest, InternalServerError, NotFound
from whoosh.searching import Hit

from abilian.core.extensions import db
from abilian.core.models.subjects import Group, User
from abilian.core.signals import activity
from abilian.core.util import utc_dt
from abilian.i18n import _, _l
from abilian.sbe.apps.communities.security import is_manager
from abilian.sbe.apps.documents.models import Document
from abilian.services.activity import ActivityEntry
from abilian.services.security import Role
from abilian.web import csrf, views
from abilian.web.action import Endpoint
from abilian.web.nav import BreadcrumbItem
from abilian.web.views import images as image_views

from ..actions import register_actions
from ..blueprint import Blueprint
from ..forms import CommunityForm
from ..models import Community, Membership
from ..security import require_admin, require_manage

__all__ = ['communities']

logger = logging.getLogger(__name__)

EPOCH = datetime.fromtimestamp(0.0, tz=pytz.utc)


def seconds_since_epoch(dt):
    if not dt:
        return 0
    return int((utc_dt(dt) - EPOCH).total_seconds())


communities = Blueprint(
    "communities",
    __name__,
    set_community_id_prefix=False,
    template_folder='../templates')
route = communities.route
add_url = communities.add_url_rule
communities.record_once(register_actions)


@communities.record_once
def register_context_processors(state):

    @state.app.context_processor
    def communities_context_processor():
        # helper to get an url for community image
        return dict(community_image_url=image_url)


def tab(tab_name):
    """
    Decorator for view functions to set the current "section" this view
    belongs to.
    """

    def decorator(f):

        @wraps(f)
        def set_current_tab(*args, **kwargs):
            g.current_tab = tab_name
            return f(*args, **kwargs)

        return set_current_tab

    return decorator


def default_view_kw(kw, obj, obj_type, obj_id, **kwargs):
    """
    Helper for using :func:`abilian.web.views.default_view` on objects that
    belongs to a community. This function should be used as `kw_func`::

        @default_view(blueprint, Model, kw_func=default_view_kw)
        @blueprint.route("/<object_id>")
        def view():
            ...

    """
    is_community = obj_type == Community.entity_type
    community_id = kw.get('community_id')

    if is_community or community_id is None:
        # when it's a community, default_view sets community_id to 'id', we want to
        # override with the slug value.
        if obj:
            if isinstance(obj, (Hit, dict)):
                community_id = obj.get('slug'
                                       if is_community else 'community_slug')
            elif is_community:
                community_id = obj.slug
            elif community_id is None and hasattr(obj, 'community'):
                try:
                    community_id = obj.community.slug
                except AttributeError:
                    pass

    if community_id is not None:
        kw['community_id'] = community_id
    else:
        raise ValueError('Cannot find community_id value')

    return kw


#
# Routes
#
@route("/")
@login_required
def index():
    query = Community.query
    sort_order = request.args.get('sort', u'').strip()
    if not sort_order:
        sort_order = session.get('sort_communities_order', 'alpha')

    if sort_order == 'activity':
        query = query.order_by(Community.last_active_at.desc())
    else:
        query = query.order_by(Community.name)

    session['sort_communities_order'] = sort_order

    if not current_user.has_role('admin'):
        # Filter with permissions
        query = query.join(Membership).filter(Membership.user == current_user)

    ctx = dict(my_communities=query.all(), sort_order=sort_order)
    return render_template("community/home.html", **ctx)


@route("/<string:community_id>/")
@views.default_view(
    communities, Community, 'community_id', kw_func=default_view_kw)
def community():
    return redirect(url_for("wall.index", community_id=g.community.slug))


@route("/json2")
def list_json2():
    """
    JSON endpoint, used for filling select boxes dynamically.
    """
    # TODO: make generic ?
    args = request.args

    q = args.get("q").replace("%", " ")
    if not q or len(q) < 2:
        raise BadRequest()

    query = db.session.query(Community.id, Community.name) \
        .filter(Community.name.ilike("%" + q + "%")) \
        .distinct() \
        .order_by(Community.name) \
        .limit(50)
    query_result = query.all()

    result = {'results': [{'id': r[0], 'text': r[1]} for r in query_result]}
    return jsonify(result)


# edit views
class BaseCommunityView(object):
    Model = Community
    pk = 'community_id'
    Form = CommunityForm
    base_template = 'community/_base.html'
    decorators = [require_admin]

    def init_object(self, args, kwargs):
        self.obj = g.community._model
        return args, kwargs

    def view_url(self):
        return url_for(self.view_endpoint, community_id=self.obj.slug)

    def get_form_kwargs(self):
        kwargs = super(BaseCommunityView, self).get_form_kwargs()

        image = self.obj.image
        if image and 'community' in g:
            setattr(image, 'url', image_url(self.obj, s=500))
            kwargs['image'] = image

        return kwargs


class CommunityEdit(BaseCommunityView, views.ObjectEdit):
    template = 'community/edit.html'
    title = _l("Edit community")
    decorators = views.ObjectEdit.decorators + (require_admin, tab('settings'))

    def breadcrumb(self):
        return BreadcrumbItem(label=_(u'Settings'),
                              icon='cog',
                              url=Endpoint('communities.settings',
                                           community_id=g.community.slug))

    def before_populate_obj(self):
        form = self.form
        name = form.name.data
        if name != self.obj.name:
            self.obj.rename(name)

        del form.name

        type = form.type.data
        if type != self.obj.type:
            self.obj.type = type
            self.obj.update_roles_on_folder()
        del form.type

        self.linked_group = form.linked_group.data or None
        if self.linked_group:
            self.linked_group = Group.query.get(int(self.linked_group))
        del form.linked_group

    def after_populate_obj(self):
        self.obj.group = self.linked_group


add_url(
    "/<string:community_id>/settings",
    view_func=CommunityEdit.as_view(
        'settings',
        view_endpoint='.community',
        message_success=_l(u"Community settings saved successfully.")))


class CommunityCreate(views.ObjectCreate, CommunityEdit):
    title = _l("Create community")
    decorators = views.ObjectCreate.decorators + (require_admin,)
    template = views.ObjectCreate.template
    base_template = views.ObjectCreate.base_template

    def breadcrumb(self):
        return BreadcrumbItem(label=_(u'Create new community'))

    def message_success(self):
        return _(u"Community %(name)s created successfully", name=self.obj.name)


add_url(
    '/new',
    view_func=CommunityCreate.as_view('new', view_endpoint='.community'))


class CommunityDelete(BaseCommunityView, views.ObjectDelete):
    get_form_kwargs = views.ObjectDelete.get_form_kwargs


add_url(
    "/<string:community_id>/destroy",
    methods=['POST'],
    view_func=CommunityDelete.as_view(
        'delete', message_success=_l(u"Community destroyed.")))

# Community Image
_DEFAULT_IMAGE = Path(__file__).parent / u'data' / u'community.png'
_DEFAULT_IMAGE_MD5 = hashlib.md5(_DEFAULT_IMAGE.open('rb').read()).hexdigest()
route('/_default_image')(image_views.StaticImageView.as_view(
    'community_default_image', set_expire=True, image=_DEFAULT_IMAGE))


class CommunityImageView(image_views.BlobView):
    id_arg = 'blob_id'

    def prepare_args(self, args, kwargs):
        community = g.community
        if not community:
            raise NotFound()

        kwargs[self.id_arg] = community.image.id
        # image = open(join(dirname(__file__), "data", "community.png"), 'rb')
        return super(CommunityImageView, self).prepare_args(args, kwargs)


image = CommunityImageView.as_view('image', max_size=500, set_expire=True)
route("/<string:community_id>/image")(image)


def image_url(community, **kwargs):
    """Return proper URL for image url."""
    if not community or not community.image:
        kwargs['md5'] = _DEFAULT_IMAGE_MD5
        return url_for('communities.community_default_image', **kwargs)

    kwargs['community_id'] = community.slug
    kwargs['md5'] = community.image.md5
    return url_for('communities.image', **kwargs)


def _members_query():
    """Helper used in members views."""
    last_activity_date = sa.sql.functions.max(ActivityEntry.happened_at) \
        .label('last_activity_date')
    memberships = User.query \
        .options(sa.orm.undefer('photo')) \
        .join(Membership) \
        .outerjoin(ActivityEntry,
                   sa.sql.and_(ActivityEntry.actor_id == User.id,
                               ActivityEntry.target_id == Membership.community_id)) \
        .filter(Membership.community == g.community, User.can_login == True) \
        .add_columns(Membership.id,
                     Membership.role,
                     last_activity_date) \
        .group_by(User, Membership.id, Membership.role) \
        .order_by(User.last_name.asc(), User.first_name.asc())

    return memberships


@route("/<string:community_id>/members")
@tab('members')
def members():
    g.breadcrumb.append(BreadcrumbItem(
        label=_(u'Members'),
        url=Endpoint('communities.members', community_id=g.community.slug))
    )
    memberships = _members_query().all()
    community_threads_users = [thread.creator for thread in g.community.threads]
    threads_count = Counter(community_threads_users)

    return render_template(
        "community/members.html",
        seconds_since_epoch=seconds_since_epoch,
        is_manager=is_manager(user=current_user),
        memberships=memberships,
        threads_count=threads_count,
        csrf_token=csrf.field())


@route("/<string:community_id>/members", methods=["POST"])
@csrf.protect
@require_manage
def members_post():
    community = g.community._model
    action = request.form.get("action")

    user_id = request.form.get("user")
    if not user_id:
        flash(_(u"You must provide a user."), 'error')
        return redirect(url_for(".members", community_id=community.slug))
    user_id = int(user_id)
    user = User.query.get(user_id)

    if action in ('add-user-role', 'set-user-role'):
        role = request.form.get("role").lower()

        community.set_membership(user, role)

        if action == 'add-user-role':
            app = current_app._get_current_object()
            activity.send(app, actor=user, verb="join", object=community)

        db.session.commit()
        return redirect(url_for('.members', community_id=community.slug))

    elif action == 'delete':
        membership_id = int(request.form['membership'])
        membership = Membership.query.get(membership_id)
        if membership.user_id != user_id:
            raise InternalServerError()

        community.remove_membership(user)

        app = current_app._get_current_object()
        activity.send(app, actor=user, verb="leave", object=community)

        db.session.commit()
        return redirect(url_for(".members", community_id=community.slug))

    else:
        raise BadRequest('Unknown action: {}'.format(repr(action)))


MEMBERS_EXPORT_HEADERS = [
    _l(u'Name'),
    _l(u'email'),
    _l(u'Last activity in this community'),
    _l(u'Role'),
]

MEMBERS_EXPORT_ATTRS = ['User', 'User.email', 'last_activity_date', 'role']

HEADER_FONT = openpyxl.styles.Font(bold=True)
HEADER_ALIGN = openpyxl.styles.Alignment(
    horizontal='center', vertical='top', wrapText=True)
XLSX_MIME = u'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'


@route("/<string:community_id>/members/excel")
@tab('members')
def members_excel_export():
    community = g.community
    attributes = [attrgetter(a) for a in MEMBERS_EXPORT_ATTRS]
    BaseModel = current_app.db.Model
    wb = openpyxl.Workbook()

    if wb.worksheets:
        wb.remove_sheet(wb.active)

    ws_title = _(u'%(community)s members', community=community.name)
    ws_title = ws_title.strip()
    if len(ws_title) > 31:
        # sheet title cannot exceed 31 char. max length
        ws_title = ws_title[:30] + u'…'
    ws = wb.create_sheet(title=ws_title)
    row = 0
    cells = []

    cols_width = []
    for col, label in enumerate(MEMBERS_EXPORT_HEADERS, 1):
        value = text_type(label)
        cell = WriteOnlyCell(ws, value=value)
        cell.font = HEADER_FONT
        cell.alignment = HEADER_ALIGN
        cells.append(cell)
        cols_width.append(len(value) + 1)

    ws.append(cells)

    for membership_info in _members_query().all():
        row += 1
        cells = []
        for col, getter in enumerate(attributes):
            value = None
            try:
                value = getter(membership_info)
            except AttributeError:
                pass

            if isinstance(value, (BaseModel, Role)):
                value = text_type(value)

            cell = WriteOnlyCell(ws, value=value)
            cells.append(value)

            # estimate width
            value = text_type(cell.value)
            width = max(len(l) for l in value.split(u'\n')) + 1
            cols_width[col] = max(width, cols_width[col])

        ws.append(cells)

    # adjust columns width
    MIN_WIDTH = 3
    MAX_WIDTH = openpyxl.utils.units.BASE_COL_WIDTH * 4

    for idx, width in enumerate(cols_width, 1):
        letter = openpyxl.utils.get_column_letter(idx)
        width = min(max(width, MIN_WIDTH), MAX_WIDTH)
        ws.column_dimensions[letter].width = width

    fd = BytesIO()
    wb.save(fd)
    fd.seek(0)

    response = current_app.response_class(fd, mimetype=XLSX_MIME)

    filename = u'{}-members-{}.xlsx'.format(
        community.slug, strftime("%d:%m:%Y-%H:%M:%S", gmtime()))
    response.headers['content-disposition'] = \
        u'attachment;filename="{}"'.format(filename)

    return response


#
# Hack to redirect from urls used by the search engine.
#
@route("/doc/<int:doc_id>")
def doc(doc_id):
    doc = Document.query.get(doc_id)

    if doc is None:
        raise NotFound()

    folder = doc.parent
    while True:
        parent = folder.parent
        if parent.is_root_folder:
            break
        folder = parent
    target_community = Community.query \
        .filter(Community.folder_id == folder.id) \
        .one()
    location = url_for(
        "documents.document_view",
        community_id=target_community.slug,
        doc_id=doc.id)
    return redirect(location)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging

import bleach
from flask import render_template_string
from jinja2 import Markup, Template

from abilian.core.util import BasePresenter
from abilian.i18n import _, _l
from abilian.sbe.apps.calendar.models import Event
from abilian.sbe.apps.communities.models import Community
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.documents.views.util import \
    breadcrumbs_for as doc_breadcrumb
from abilian.sbe.apps.forum.models import Post, Thread
from abilian.sbe.apps.wiki.models import WikiPage
from abilian.web.util import url_for

logger = logging.getLogger(__name__)

# Poor man's pattern matching.
MESSAGES = {
    ('post', Thread): _l('has started conversation "{object}"'),
    ('post', Post): _l('has participated in conversation "{object}"'),
    ('post', Document): _l('has published document "{object}"'),
    ('post', WikiPage): _l('has created wiki page "{object}"'),
    ('post', Event): _l('has published event "{object}"'),
    ('post', Community): _l('has created communauté "{object}"'),
    #
    ('update', Document): _l('has updated document "{object}"'),
    ('update', WikiPage): _l('has updated wiki page "{object}"'),
    ('update', Community): _l('has updated community "{object}"'),
    #
    ('join', Community): _l('has joined the community {object}.'),
    #
    ('leave', Community): _l('has left the community {object}.'),
}

OBJ_TEMPLATE = Template('<a href="{{ object_url|safe }}">{{ object_name }}</a>')

POST_BODY_TEMPLATE = '''
  <span class="body">"<a href="{{ object_url |safe }}">{{ body }}</a>"
  {%- if post.attachments %}
  <div id="attachments">
    <ul>
      {%- for attachment in post.attachments %}
      <li>
        <span class="attachment-item">
          <img src="{{ attachment.icon }}""
              alt=""/>
          <a href="{{ url_for(attachment) }}">{{ attachment.name }}</a>
          ({{ attachment.content_length|filesize }})
        </span>
      </li>
      {%- endfor %}
    </ul>
  </div>
  {%- endif %}
  </span>
 '''

DOCUMENT_BODY_TEMPLATE = '''
<div class="body">
  <div>
    <img src="{{ obj.icon }}" style="vertical-align: middle;" alt=""/>
    {% for p in parents[:-1] %}
    <a href="{{ p.path }}">{{ p.label }}</a> <span class="divider">/</span>
    {% endfor %}

    <a href="{{ url_for(obj) }}">{{ obj.name }}</a>
  </div>
  <div>
  {%- if obj.antivirus_ok %}
    <a href="{{ url_for('documents.document_download',
                         community_id=obj.community.slug,
                         doc_id=obj.id,
                         attach=True) }}">
      <i class="glyphicon glyphicon-download"></i>
      {{ _('Download') }} (<small>{{ obj.content_length|filesize  }}</small>)
    </a>
  {%- endif %}
  </div>
</div>
'''


class ActivityEntryPresenter(BasePresenter):

    @property
    def object_url(self):
        return url_for(self._model.object)

    def message(self, ignore_community=False):
        try:
            # another quick&dirty approach for now. FIXME later.
            entry = self._model
            object_class = entry.object_type.split('.')[-1]
            object_class_localized = _(object_class)

            ctx = {}
            ctx['verb'] = entry.verb

            ctx['object_name'] = entry.object.name or getattr(entry.object,
                                                              'title', "???")
            ctx['object_url'] = url_for(entry.object)
            ctx['object_type'] = object_class_localized
            ctx['object'] = OBJ_TEMPLATE.render(**ctx)

            if entry.target:
                ctx['target_name'] = entry.target.name
                ctx['target_url'] = url_for(entry.target)
                ctx['target'] = OBJ_TEMPLATE.render(
                    object_name=ctx['target_name'],
                    object_url=ctx['target_url'])

            msg = MESSAGES.get((entry.verb, entry.object.__class__))
            if msg:
                msg = msg.format(**ctx)
                if entry.target and not ignore_community:
                    msg += " " + _('in the community {target}.').format(**ctx)
                else:
                    msg += "."

            elif entry.verb == 'post':
                msg = _('has posted an object of type {object_type} '
                        'called "{object}"').format(**ctx)

                if entry.target and not ignore_community:
                    msg += " " + _('in the community {target}.').format(**ctx)
                else:
                    msg += "."

            elif entry.verb == 'join':
                msg = _('has joined the community {object}.').format(**ctx)

            elif entry.verb == 'leave':
                msg = _('has left the community {object}.').format(**ctx)

            elif entry.verb == 'update':
                msg = _('has updated {object_type} {object}.').format(**ctx)

            else:
                msg = _('has done action "{verb}" on object "{object}".').format(**ctx)

            return Markup(msg)

        except:
            logger.exception('Exception while presenting activity message')
            raise

    @property
    def body(self):
        if isinstance(self.object, Thread):
            body = bleach.clean(
                self.object.posts[0].body_html, tags=[], strip=True)
            body = Markup(body).unescape()
            if len(body) > 400:
                body = body[0:400] + "…"
            body = render_template_string(
                POST_BODY_TEMPLATE,
                object_url=self.object_url,
                body=body,
                post=self.object.posts[0])
            return Markup(body)
        elif isinstance(self.object, Post):
            body = bleach.clean(self.object.body_html, tags=[], strip=True)
            body = Markup(body).unescape()
            if len(body) > 400:
                body = body[0:400] + "…"
            body = render_template_string(
                POST_BODY_TEMPLATE,
                object_url=self.object_url,
                body=body,
                post=self.object)
            return Markup(body)
        elif isinstance(self.object, Document):
            parents = doc_breadcrumb(self.object)
            body = render_template_string(
                DOCUMENT_BODY_TEMPLATE, obj=self.object, parents=parents)
            return Markup(body)
        else:
            return ""



# coding=utf-8

from __future__ import absolute_import, print_function

from flask import g, request, url_for
from flask_login import current_user

from abilian.i18n import _l
from abilian.services import get_service
from abilian.services.security import Admin
from abilian.web.action import Action, FAIcon, ModalActionMixin, actions


class ForumAction(Action):

    def is_current(self):
        return request.path == self.url()

    def is_filtered(self):
        filter_keys = ["today", "month", "year", "week"]
        current_url = request.path.split("/")
        filter = current_url[-1].strip()

        if filter not in filter_keys:
            return False

        if self.url() == "#filter":
            return True

        return False

    def url(self, context=None):
        if self._url or self.endpoint:
            return super(ForumAction, self).url(context=context)

        return url_for("." + self.name, community_id=g.community.slug)


class ThreadAction(ForumAction):

    def pre_condition(self, context):
        thread = actions.context.get('object')
        return not not thread


def is_admin(context):
    security = get_service('security')
    return security.has_role(current_user, Admin, object=context.get('object'))


def is_in_thread(context):
    thread = context.get('object')
    return not thread


def is_closed(context):
    thread = context.get('object')
    return thread.closed


def not_closed(context):
    return not is_closed(context)


class ForumModalAction(ModalActionMixin, ThreadAction):
    pass


_close_template_action = u'''
<form method="POST" action="{{ url }}" encoding="multipart/form-data">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
'''

_actions = (
    ForumAction('forum:global', 'index', _l(u'Recent conversations')),
    ForumAction(
        'forum:global', 'index/<string:filter>', _l(u'Top'), url='#filter'),
    ForumAction('forum:global', 'archives', _l(u'Archives')),
    ForumAction(
        'forum:global',
        'attachments',
        _l(u'Attachments'),
        condition=is_in_thread),
    ForumAction(
        'forum:global', 'new_thread', _l(u'New conversation'), icon='plus'),
    ForumModalAction(
        'forum:thread',
        'delete',
        _l(u'Delete'),
        condition=lambda ctx: is_admin(ctx) and not_closed(ctx),
        url='#modal-delete',
        icon='trash'),
    #
    ThreadAction(
        'forum:thread',
        'close',
        _l('Close thread'),
        url='close',
        template_string=_close_template_action,
        condition=lambda ctx: is_admin(ctx) and not_closed(ctx),
        icon=FAIcon('lock')),
    ThreadAction(
        'forum:thread',
        'reopen',
        _l('Re-open thread'),
        url='close',
        template_string=_close_template_action,
        condition=lambda ctx: is_admin(ctx) and is_closed(ctx),
        icon=FAIcon('unlock')),
    ThreadAction(
        'forum:thread',
        'attachments',
        _l(u'Attachments'),
        url='attachments',
        icon='file'),)


def register_actions(state):
    if not actions.installed(state.app):
        return
    with state.app.app_context():
        actions.register(*_actions)



from __future__ import absolute_import, print_function

from flask import url_for

from abilian.core.models.subjects import User
from abilian.sbe.testing import BaseTestCase


class TestSocialViews(BaseTestCase):

    init_data = True
    no_login = True

    # Tests start here
    def test_home(self):
        response = self.client.get(url_for("social.home"))
        self.assert_200(response)
        self.assert_valid(response)

    def test_users(self):
        response = self.client.get(url_for("social.users"))
        self.assert_200(response)
        self.assert_valid(response)

    def test_groups(self):
        response = self.client.get(url_for("social.groups"))
        self.assert_200(response)
        self.assert_valid(response)

    def test_user(self):
        user = User.query.all()[0]
        response = self.client.get(url_for("social.user", user_id=user.id))
        self.assert_200(response)
        self.assert_valid(response)



# coding=utf-8

from __future__ import absolute_import, print_function

from unittest import TestCase

from six import text_type

from abilian.core.entities import all_entity_classes
from abilian.core.models.subjects import Group, User

from ..models import Message, PrivateMessage


def check_editable(object):
    if hasattr(object, '__editable__'):
        for k in object.__editable__:
            assert hasattr(object, k)


class TestUsers(TestCase):

    def test_user(self):
        user = User(
            first_name=u"John",
            last_name=u"Test User",
            email=u"test@example.com",
            password="toto")
        check_editable(user)

        assert u"John Test User" == user.name
        assert u"John Test User" == text_type(user)
        # self.assertEquals(len(user.messages), 0)

    def test_user_follow(self):
        # id is provided by DB (which is not used in this test), and is required for
        # having in (user1 != user2) == True
        user1 = User(
            id=1,
            first_name=u"John",
            last_name=u"Test User 1",
            email=u"test1@example.com",
            password="toto")
        user2 = User(
            id=2,
            first_name=u"Joe",
            last_name=u"Test User 2",
            email=u"test2@example.com",
            password="toto")

        assert len(user1.followers) == 0
        assert len(user1.followees) == 0
        assert len(user2.followers) == 0
        assert len(user2.followees) == 0

        user1.follow(user2)

        assert user1.is_following(user2)
        assert len(user2.followers) == 1
        assert len(user1.followees) == 1
        assert len(user2.followees) == 0
        assert len(user1.followers) == 0
        assert user2 in user1.followees

        user1.unfollow(user2)

        assert not user1.is_following(user2)
        assert len(user1.followers) == 0
        assert len(user1.followees) == 0
        assert len(user2.followers) == 0
        assert len(user2.followees) == 0
        assert user2 not in user1.followers

    def test_group(self):
        user = User(
            first_name=u"John",
            last_name=u"Test User",
            email=u"test@example.com",
            password="toto")
        group = Group(name=u"Group 1")

        user.join(group)

        assert user.is_member_of(group)
        assert len(group.members) == 1
        assert len(user.groups) == 1
        assert group.members == {user}
        assert user.groups == {group}

        user.leave(group)

        assert not user.is_member_of(group)
        assert len(group.members) == 0
        assert len(user.groups) == 0


class TestContent(TestCase):

    def test_private_message(self):
        pm = PrivateMessage(creator_id=0, recipient_id=0)
        check_editable(pm)

    def test_status_update(self):
        user = User(
            first_name=u"John",
            last_name=u"Test User",
            email=u"test@example.com",
            password="toto")
        # self.assertEquals(len(user.messages), 0)

        message = Message()
        message.author = user
        check_editable(message)

        # self.assertEquals(len(user.messages), 1)
        # self.assertEquals(user.messages[0], message)
        # self.assertEquals(message.author_id, user.uid)

    def test_tags(self):
        m = Message(content=u"abc #123 #cde #voilà_l_été #789")
        assert m.tags == [u'123', u'cde', u'voilà_l_été', u'789']

    def test_get_all_entity_classes(self):
        classes = all_entity_classes()
        assert Message in classes
        assert PrivateMessage in classes



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function

import logging
import pkgutil
from cgi import escape

import sqlalchemy as sa
from flask import Response, flash, g, jsonify, redirect, render_template, \
    request
from six import text_type
from sqlalchemy.sql.expression import and_, asc, desc, func, nullslast, or_
from werkzeug.exceptions import InternalServerError

from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.i18n import _, _l
from abilian.sbe.apps.communities.models import Membership
from abilian.sbe.apps.wall.presenters import ActivityEntryPresenter
from abilian.sbe.apps.wall.views import get_recent_entries
from abilian.services import get_service
from abilian.web import url_for
from abilian.web.filters import age
from abilian.web.views import ObjectEdit, default_view
from abilian.web.views.images import user_photo_url

from ..forms import UserProfileForm, UserProfileViewForm
from .social import social
from .util import Env

logger = logging.getLogger(__name__)

DEFAULT_USER_MUGSHOT = pkgutil.get_data('abilian.sbe',
                                        'static/images/silhouette_unknown.png')


def make_tabs(user):
    return [
        dict(id='profile', label=_(u'Profile'), link=url_for(user, tab='profile')),
        # dict(id='conversations', label=_(u'Conversations'), link=url_for(user), is_online=True),
        dict(id='documents', label=_(u'Documents'), link=url_for(user, tab='documents')),
        dict(id='images', label=_(u'Images'), link=url_for(user, tab='images')),
        dict(id='audit', label=_(u'Audit'), link=url_for(user, tab='audit')),
    ]


@social.route("/users/")
def users():
    query = request.args.get("query")

    if query:
        query = query.replace("%", " ")
        q = or_(
            User.first_name.like("%" + query + "%"),
            User.last_name.like("%" + query + "%"))
        users = User.query.filter(q).limit(100).all()
    else:
        users = User.query.limit(100).all()
    ctx = dict(users=users)
    return render_template("social/users.html", **ctx)


@social.route("/users/dt_json")
def users_dt_json():
    """JSON call to fill a DataTable. Needs some refactoring."""
    args = request.args

    length = int(args.get("iDisplayLength", 0))
    start = int(args.get("iDisplayStart", 0))
    sort_col = int(args.get("iSortCol_0", 1))
    sort_dir = args.get("sSortDir_0", "asc")
    echo = int(args.get("sEcho", 0))
    search = args.get("sSearch", "").replace("%", "").lower()

    end = start + length

    query = User.query
    total_count = query.count()

    if search:
        # TODO: gérer les accents
        filter = or_(
            func.lower(User.first_name).like("%" + search + "%"),
            func.lower(User.last_name).like("%" + search + "%"))
        query = query.filter(filter) \
            .reset_joinpoint()

    count = query.count()
    SORT_COLS = {
        1: [],  # will be set to [User.last_name, User.first_name]
        2: [User.created_at],
        3: [User.last_active],
    }
    columns = list(SORT_COLS.get(sort_col, []))
    columns.extend([func.lower(User.last_name), func.lower(User.first_name)])

    direction = asc if sort_dir == 'asc' else desc
    order_by = map(direction, columns)

    # sqlite does not support 'NULLS FIRST|LAST' in ORDER BY clauses
    engine = query.session.get_bind(User.__mapper__)
    if engine.name != 'sqlite':
        order_by[0] = nullslast(order_by[0])

    query = query.order_by(*order_by)

    users = query.slice(start, end).all()

    data = []
    MUGSHOT_SIZE = 45
    for user in users:
        # TODO: this should be done on the browser.
        user_url = url_for(".user", user_id=user.id)
        mugshot = user_photo_url(user, size=MUGSHOT_SIZE)
        name = escape(getattr(user, "name") or "")

        cell0 = (
            u'<a href="{url}"><img src="{src}" width="{size}" height="{size}">'
            '</a>'.format(url=user_url, src=mugshot, size=MUGSHOT_SIZE))
        cell1 = (u'<div class="info"><a href="{user_url}">{name}</a> '
                 u'</div>'.format(**locals()))
        cell2 = age(user.created_at)
        cell3 = age(user.last_active)

        cell4 = ''  # TODO: follow / unfollow?
        data.append([cell0, cell1, cell2, cell3])

    result = {
        "sEcho": echo,
        "iTotalRecords": total_count,
        "iTotalDisplayRecords": count,
        "aaData": data,
    }
    return jsonify(result)


@social.route("/users/<int:user_id>")
@default_view(social, User, 'user_id')
def user(user_id):
    security = get_service('security')
    user = User.query.get(user_id)

    # FIXME: use user profiles
    view_form = UserProfileViewForm(obj=user)

    communities = [m.community for m in user.communautes_membership]
    if g.user != user and (not security.has_role(g.user, 'manager')):
        # filter visible communautes (ticket 165)
        communities = [c for c in communities if c.has_member(g.user)]

    view_form.communautes._set_data(communities)

    # FIXME
    contact = user
    env = Env(user=user,
              contact=contact,
              view_form=view_form,
              can_edit=can_edit(user),
              tabs=make_tabs(user))

    entries = get_recent_entries(user=user)
    entries = ActivityEntryPresenter.wrap_collection(entries)
    env.activity_entries = entries

    return render_template("social/user.html", **env)


def can_edit(user):
    security = get_service('security')
    if not security:
        return True

    # TODO: introduce a "self" role?
    return (user == g.user) or security.has_role(g.user, 'admin')


class UserProfileEdit(ObjectEdit):

    Model = User
    Form = UserProfileForm
    pk = 'user_id'
    _message_success = _l(u'Profile edited')

    def init_object(self, args, kwargs):
        args, kwargs = super(UserProfileEdit, self).init_object(args, kwargs)
        self.user = self.obj
        return args, kwargs

    def view_url(self):
        return url_for('.user', user_id=self.user.id)

    def edit(self):
        if not can_edit(self.user):
            return Response(status=403)
        return super(UserProfileEdit, self).edit()

    def handle_commit_exception(self, exc):
        db.session.rollback()
        if isinstance(exc, sa.exc.IntegrityError):
            log_msg = 'Error saving user profile'
        else:
            log_msg = 'Unexpected error while saving user profile'
        logger.error(log_msg, exc_info=True, extra={'stack': True})
        flash(_(u'Error occured'), "error")
        return self.redirect_to_view()


social.route("/users/<int:user_id>/edit")(UserProfileEdit.as_view('user_edit'))


@social.route("/users/<int:user_id>", methods=['POST'])
def user_post(user_id):
    user = User.query.get(user_id)
    action = request.form.get('action')
    return_url = request.form.get('return_url')

    if action == 'follow':
        g.user.follow(user)
    elif action == 'unfollow':
        g.user.unfollow(user)
    else:
        raise Exception("Should not happen")
    db.session.commit()

    if return_url:
        # TODO: security check
        return redirect(return_url)
    else:
        return redirect(url_for(".user_view", user_id=user_id))


#
# Ad-hoc JSON endpoints, used by select boxes
#
@social.route("/users/json")
def users_json():
    q = request.args.get("q").replace("%", " ").lower()

    if not q or len(q) < 2:
        raise InternalServerError()

    query = User.query \
        .filter(or_(func.lower(User.first_name).like(q + "%"),
                    func.lower(User.last_name).like(q + "%"))) \
        .order_by(func.lower(User.last_name))

    with_membership = request.args.get('with_membership')
    if with_membership is not None:
        # provide membership info for a community
        with_membership = int(with_membership)
        query = query \
            .outerjoin(Membership,
                       and_(Membership.user.expression,
                            Membership.community_id == with_membership)) \
            .filter(User.can_login == True) \
            .add_columns(Membership.role)

    exclude_community = request.args.get('exclude_community')
    if exclude_community is not None:
        exclude_community = int(exclude_community)
        exclude = ~Membership.query \
            .filter(Membership.user.expression,
                    Membership.community_id == exclude_community) \
            .options(sa.orm.noload('user'),
                     sa.orm.noload('community')) \
            .exists()
        query = query.filter(exclude)

    results = []
    for user in query.all():
        role = None
        if with_membership is not None:
            user, role = user

        if role is not None:
            role = text_type(role)

        item = {
            'id': user.id,
            'text': u'{} ({})'.format(user.name, user.email),
            'name': user.name,
            'email': user.email,
            'role': role,
        }
        results.append(item)

    return jsonify({'results': results})
